"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const semaphore_async_await_1 = __importDefault(require("semaphore-async-await"));
const ethereumjs_util_1 = require("ethereumjs-util");
const block_1 = require("@ethereumjs/block");
const ethash_1 = __importDefault(require("@ethereumjs/ethash"));
const common_1 = __importDefault(require("@ethereumjs/common"));
const manager_1 = require("./db/manager");
const helpers_1 = require("./db/helpers");
const operation_1 = require("./db/operation");
const level = require('level-mem');
/**
 * This class stores and interacts with blocks.
 */
class Blockchain {
    /**
     * Creates new Blockchain object
     *
     * @param opts - An object with the options that this constructor takes. See
     * [[BlockchainOptions]].
     */
    constructor(opts = {}) {
        var _a, _b;
        // Throw on chain or hardfork options removed in latest major release to
        // prevent implicit chain setup on a wrong chain
        if ('chain' in opts || 'hardfork' in opts) {
            throw new Error('Chain/hardfork options are not allowed any more on initialization');
        }
        if (opts.common) {
            this._common = opts.common;
        }
        else {
            const DEFAULT_CHAIN = 'mainnet';
            const DEFAULT_HARDFORK = 'chainstart';
            this._common = new common_1.default({
                chain: DEFAULT_CHAIN,
                hardfork: DEFAULT_HARDFORK,
            });
        }
        this._validateConsensus = (_a = opts.validateConsensus) !== null && _a !== void 0 ? _a : true;
        this._validateBlocks = (_b = opts.validateBlocks) !== null && _b !== void 0 ? _b : true;
        this.db = opts.db ? opts.db : level();
        this.dbManager = new manager_1.DBManager(this.db, this._common);
        if (this._validateConsensus) {
            if (this._common.consensusType() !== 'pow') {
                throw new Error('consensus validation only supported for pow chains');
            }
            if (this._common.consensusAlgorithm() !== 'ethash') {
                throw new Error('consensus validation only supported for pow ethash algorithm');
            }
            this._ethash = new ethash_1.default(this.db);
        }
        this._heads = {};
        this._lock = new semaphore_async_await_1.default(1);
        if (opts.genesisBlock && !opts.genesisBlock.isGenesis()) {
            throw 'supplied block is not a genesis block';
        }
        // eslint-disable-next-line @typescript-eslint/no-floating-promises
        this.initPromise = this._init(opts.genesisBlock);
    }
    /**
     * This static constructor safely creates a new Blockchain object, which also
     * awaits the initialization function. If this initialization function throws,
     * then this constructor will throw as well, and is therefore the encouraged
     * method to use when creating a blockchain object.
     * @param opts Constructor options, see [[BlockchainOptions]]
     */
    static async create(opts = {}) {
        const blockchain = new Blockchain(opts);
        await blockchain.initPromise.catch((e) => {
            throw e;
        });
        return blockchain;
    }
    /**
     * Returns an object with metadata about the Blockchain. It's defined for
     * backwards compatibility.
     */
    get meta() {
        return {
            rawHead: this._headHeaderHash,
            heads: this._heads,
            genesis: this._genesis,
        };
    }
    /**
     * This method is called in the constructor and either sets up the DB or reads
     * values from the DB and makes these available to the consumers of
     * Blockchain.
     *
     * @hidden
     */
    async _init(genesisBlock) {
        let DBGenesisBlock;
        try {
            const genesisHash = await this.dbManager.numberToHash(new ethereumjs_util_1.BN(0));
            DBGenesisBlock = await this.dbManager.getBlock(genesisHash);
        }
        catch (error) {
            if (error.type !== 'NotFoundError') {
                throw error;
            }
        }
        if (!genesisBlock) {
            const common = new common_1.default({
                chain: this._common.chainId(),
                hardfork: 'chainstart',
            });
            genesisBlock = block_1.Block.genesis({}, { common });
        }
        // If the DB has a genesis block, then verify that the genesis block in the
        // DB is indeed the Genesis block generated or assigned.
        if (DBGenesisBlock && !genesisBlock.hash().equals(DBGenesisBlock.hash())) {
            throw new Error('The genesis block in the DB has a different hash than the provided genesis block.');
        }
        const genesisHash = genesisBlock.hash();
        if (!DBGenesisBlock) {
            // If there is no genesis block put the genesis block in the DB.
            // For that TD, the BlockOrHeader, and the Lookups have to be saved.
            const dbOps = [];
            dbOps.push(helpers_1.DBSetTD(genesisBlock.header.difficulty.clone(), new ethereumjs_util_1.BN(0), genesisHash));
            helpers_1.DBSetBlockOrHeader(genesisBlock).map((op) => dbOps.push(op));
            helpers_1.DBSaveLookups(genesisHash, new ethereumjs_util_1.BN(0)).map((op) => dbOps.push(op));
            await this.dbManager.batch(dbOps);
        }
        // At this point, we can safely set genesisHash as the _genesis hash in this
        // object: it is either the one we put in the DB, or it is equal to the one
        // which we read from the DB.
        this._genesis = genesisHash;
        // load verified iterator heads
        try {
            const heads = await this.dbManager.getHeads();
            this._heads = heads;
        }
        catch (error) {
            if (error.type !== 'NotFoundError') {
                throw error;
            }
            this._heads = {};
        }
        // load headerchain head
        try {
            const hash = await this.dbManager.getHeadHeader();
            this._headHeaderHash = hash;
        }
        catch (error) {
            if (error.type !== 'NotFoundError') {
                throw error;
            }
            this._headHeaderHash = genesisHash;
        }
        // load blockchain head
        try {
            const hash = await this.dbManager.getHeadBlock();
            this._headBlockHash = hash;
        }
        catch (error) {
            if (error.type !== 'NotFoundError') {
                throw error;
            }
            this._headBlockHash = genesisHash;
        }
    }
    /**
     * Perform the `action` function after we have initialized this module and
     * have acquired a lock
     * @param action - the action function to run after initializing and acquiring
     * a lock
     * @hidden
     */
    async initAndLock(action) {
        await this.initPromise;
        return await this.runWithLock(action);
    }
    /**
     * Run a function after acquiring a lock. It is implied that we have already
     * initialized the module (or we are calling this from the init function, like
     * `_setCanonicalGenesisBlock`)
     * @param action - function to run after acquiring a lock
     * @hidden
     */
    async runWithLock(action) {
        try {
            await this._lock.acquire();
            const value = await action();
            return value;
        }
        finally {
            this._lock.release();
        }
    }
    /**
     * Returns the specified iterator head.
     *
     * @param name - Optional name of the state root head (default: 'vm')
     */
    async getHead(name = 'vm') {
        return await this.initAndLock(async () => {
            // if the head is not found return the headHeader
            const hash = this._heads[name] || this._headBlockHash;
            if (!hash) {
                throw new Error('No head found.');
            }
            const block = await this._getBlock(hash);
            return block;
        });
    }
    /**
     * Returns the latest header in the canonical chain.
     */
    async getLatestHeader() {
        return await this.initAndLock(async () => {
            if (!this._headHeaderHash) {
                throw new Error('No head header set');
            }
            const block = await this._getBlock(this._headHeaderHash);
            return block.header;
        });
    }
    /**
     * Returns the latest full block in the canonical chain.
     */
    async getLatestBlock() {
        return this.initAndLock(async () => {
            if (!this._headBlockHash) {
                throw new Error('No head block set');
            }
            const block = this._getBlock(this._headBlockHash);
            return block;
        });
    }
    /**
     * Adds blocks to the blockchain.
     *
     * If an invalid block is met the function will throw, blocks before will
     * nevertheless remain in the DB. If any of the saved blocks has a higher
     * total difficulty than the current max total difficulty the canonical
     * chain is rebuilt and any stale heads/hashes are overwritten.
     * @param blocks - The blocks to be added to the blockchain
     */
    async putBlocks(blocks) {
        await this.initPromise;
        for (let i = 0; i < blocks.length; i++) {
            await this.putBlock(blocks[i]);
        }
    }
    /**
     * Adds a block to the blockchain.
     *
     * If the block is valid and has a higher total difficulty than the current
     * max total difficulty, the canonical chain is rebuilt and any stale
     * heads/hashes are overwritten.
     * @param block - The block to be added to the blockchain
     */
    async putBlock(block) {
        await this.initPromise;
        await this._putBlockOrHeader(block);
    }
    /**
     * Adds many headers to the blockchain.
     *
     * If an invalid header is met the function will throw, headers before will
     * nevertheless remain in the DB. If any of the saved headers has a higher
     * total difficulty than the current max total difficulty the canonical
     * chain is rebuilt and any stale heads/hashes are overwritten.
     * @param headers - The headers to be added to the blockchain
     */
    async putHeaders(headers) {
        await this.initPromise;
        for (let i = 0; i < headers.length; i++) {
            await this.putHeader(headers[i]);
        }
    }
    /**
     * Adds a header to the blockchain.
     *
     * If this header is valid and it has a higher total difficulty than the current
     * max total difficulty, the canonical chain is rebuilt and any stale
     * heads/hashes are overwritten.
     * @param header - The header to be added to the blockchain
     */
    async putHeader(header) {
        await this.initPromise;
        await this._putBlockOrHeader(header);
    }
    /**
     * Entrypoint for putting any block or block header. Verifies this block,
     * checks the total TD: if this TD is higher than the current highest TD, we
     * have thus found a new canonical block and have to rewrite the canonical
     * chain. This also updates the head block hashes. If any of the older known
     * canonical chains just became stale, then we also reset every _heads header
     * which points to a stale header to the last verified header which was in the
     * old canonical chain, but also in the new canonical chain. This thus rolls
     * back these headers so that these can be updated to the "new" canonical
     * header using the iterator method.
     * @hidden
     */
    async _putBlockOrHeader(item) {
        await this.runWithLock(async () => {
            const block = item instanceof block_1.BlockHeader ? new block_1.Block(item) : item;
            const isGenesis = block.isGenesis();
            // we cannot overwrite the Genesis block after initializing the Blockchain
            if (isGenesis) {
                throw new Error('Cannot put a genesis block: create a new Blockchain');
            }
            const { header } = block;
            const blockHash = header.hash();
            const blockNumber = header.number;
            const td = header.difficulty.clone();
            const currentTd = { header: new ethereumjs_util_1.BN(0), block: new ethereumjs_util_1.BN(0) };
            let dbOps = [];
            if (block._common.chainId() !== this._common.chainId()) {
                throw new Error('Chain mismatch while trying to put block or header');
            }
            if (this._validateBlocks && !isGenesis) {
                // this calls into `getBlock`, which is why we cannot lock yet
                await block.validate(this);
            }
            if (this._validateConsensus) {
                if (this._ethash) {
                    const valid = await this._ethash.verifyPOW(block);
                    if (!valid) {
                        throw new Error('invalid POW');
                    }
                }
            }
            // set total difficulty in the current context scope
            if (this._headHeaderHash) {
                currentTd.header = await this.getTotalDifficulty(this._headHeaderHash);
            }
            if (this._headBlockHash) {
                currentTd.block = await this.getTotalDifficulty(this._headBlockHash);
            }
            // calculate the total difficulty of the new block
            const parentTd = await this.getTotalDifficulty(header.parentHash, blockNumber.subn(1));
            td.iadd(parentTd);
            // save block and total difficulty to the database
            dbOps = dbOps.concat(helpers_1.DBSetTD(td, blockNumber, blockHash));
            // save header/block
            dbOps = dbOps.concat(helpers_1.DBSetBlockOrHeader(block));
            // if total difficulty is higher than current, add it to canonical chain
            if (block.isGenesis() || td.gt(currentTd.header)) {
                this._headHeaderHash = blockHash;
                if (item instanceof block_1.Block) {
                    this._headBlockHash = blockHash;
                }
                // TODO SET THIS IN CONSTRUCTOR
                if (block.isGenesis()) {
                    this._genesis = blockHash;
                }
                // delete higher number assignments and overwrite stale canonical chain
                await this._deleteCanonicalChainReferences(blockNumber.addn(1), blockHash, dbOps);
                // from the current header block, check the blockchain in reverse (i.e.
                // traverse `parentHash`) until `numberToHash` matches the current
                // number/hash in the canonical chain also: overwrite any heads if these
                // heads are stale in `_heads` and `_headBlockHash`
                await this._rebuildCanonical(header, dbOps);
            }
            else {
                // the TD is lower than the current highest TD so we will add the block
                // to the DB, but will not mark it as the canonical chain.
                if (td.gt(currentTd.block) && item instanceof block_1.Block) {
                    this._headBlockHash = blockHash;
                }
                // save hash to number lookup info even if rebuild not needed
                dbOps.push(helpers_1.DBSetHashToNumber(blockHash, blockNumber));
            }
            const ops = dbOps.concat(this._saveHeadOps());
            await this.dbManager.batch(ops);
        });
    }
    /**
     * Gets a block by its hash.
     *
     * @param blockId - The block's hash or number. If a hash is provided, then
     * this will be immediately looked up, otherwise it will wait until we have
     * unlocked the DB
     */
    async getBlock(blockId) {
        // cannot wait for a lock here: it is used both in `validate` of `Block`
        // (calls `getBlock` to get `parentHash`) it is also called from `runBlock`
        // in the `VM` if we encounter a `BLOCKHASH` opcode: then a BN is used we
        // need to then read the block from the canonical chain Q: is this safe? We
        // know it is OK if we call it from the iterator... (runBlock)
        await this.initPromise;
        return await this._getBlock(blockId);
    }
    /**
     * @hidden
     */
    async _getBlock(blockId) {
        return this.dbManager.getBlock(blockId);
    }
    /**
     * Gets total difficulty for a block specified by hash and number
     */
    async getTotalDifficulty(hash, number) {
        if (!number) {
            number = await this.dbManager.hashToNumber(hash);
        }
        return this.dbManager.getTotalDifficulty(hash, number);
    }
    /**
     * Looks up many blocks relative to blockId Note: due to `GetBlockHeaders
     * (0x03)` (ETH wire protocol) we have to support skip/reverse as well.
     * @param blockId - The block's hash or number
     * @param maxBlocks - Max number of blocks to return
     * @param skip - Number of blocks to skip apart
     * @param reverse - Fetch blocks in reverse
     */
    async getBlocks(blockId, maxBlocks, skip, reverse) {
        return await this.initAndLock(async () => {
            const blocks = [];
            let i = -1;
            const nextBlock = async (blockId) => {
                let block;
                try {
                    block = await this._getBlock(blockId);
                }
                catch (error) {
                    if (error.type !== 'NotFoundError') {
                        throw error;
                    }
                    return;
                }
                i++;
                const nextBlockNumber = block.header.number.addn(reverse ? -1 : 1);
                if (i !== 0 && skip && i % (skip + 1) !== 0) {
                    return await nextBlock(nextBlockNumber);
                }
                blocks.push(block);
                if (blocks.length < maxBlocks) {
                    await nextBlock(nextBlockNumber);
                }
            };
            await nextBlock(blockId);
            return blocks;
        });
    }
    /**
     * Given an ordered array, returns an array of hashes that are not in the
     * blockchain yet. Uses binary search to find out what hashes are missing.
     * Therefore, the array needs to be ordered upon number.
     * @param hashes - Ordered array of hashes (ordered on `number`).
     */
    async selectNeededHashes(hashes) {
        return await this.initAndLock(async () => {
            let max;
            let mid;
            let min;
            max = hashes.length - 1;
            mid = min = 0;
            while (max >= min) {
                let number;
                try {
                    number = await this.dbManager.hashToNumber(hashes[mid]);
                }
                catch (error) {
                    if (error.type !== 'NotFoundError') {
                        throw error;
                    }
                }
                if (number) {
                    min = mid + 1;
                }
                else {
                    max = mid - 1;
                }
                mid = Math.floor((min + max) / 2);
            }
            return hashes.slice(min);
        });
    }
    /**
     * Completely deletes a block from the blockchain including any references to
     * this block. If this block was in the canonical chain, then also each child
     * block of this block is deleted Also, if this was a canonical block, each
     * head header which is part of this now stale chain will be set to the
     * parentHeader of this block An example reason to execute is when running the
     * block in the VM invalidates this block: this will then reset the canonical
     * head to the past block (which has been validated in the past by the VM, so
     * we can be sure it is correct).
     * @param blockHash - The hash of the block to be deleted
     */
    async delBlock(blockHash) {
        // Q: is it safe to make this not wait for a lock? this is called from
        // `runBlockchain` in case `runBlock` throws (i.e. the block is invalid).
        // But is this the way to go? If we know this is called from the
        // iterator/runBlockchain we are safe, but if this is called from anywhere
        // else then this might lead to a concurrency problem?
        await this.initPromise;
        await this._delBlock(blockHash);
    }
    /**
     * @hidden
     */
    async _delBlock(blockHash) {
        const dbOps = [];
        // get header
        const header = await this._getHeader(blockHash);
        const blockHeader = header;
        const blockNumber = blockHeader.number;
        const parentHash = blockHeader.parentHash;
        // check if block is in the canonical chain
        const canonicalHash = await this.safeNumberToHash(blockNumber);
        const inCanonical = !!canonicalHash && canonicalHash.equals(blockHash);
        // delete the block, and if block is in the canonical chain, delete all
        // children as well
        await this._delChild(blockHash, blockNumber, inCanonical ? parentHash : null, dbOps);
        // delete all number to hash mappings for deleted block number and above
        if (inCanonical) {
            await this._deleteCanonicalChainReferences(blockNumber, parentHash, dbOps);
        }
        await this.dbManager.batch(dbOps);
    }
    /**
     * Updates the `DatabaseOperation` list to delete a block from the DB,
     * identified by `blockHash` and `blockNumber`. Deletes fields from `Header`,
     * `Body`, `HashToNumber` and `TotalDifficulty` tables. If child blocks of
     * this current block are in the canonical chain, delete these as well. Does
     * not actually commit these changes to the DB. Sets `_headHeaderHash` and
     * `_headBlockHash` to `headHash` if any of these matches the current child to
     * be deleted.
     * @param blockHash - the block hash to delete
     * @param blockNumber - the number corresponding to the block hash
     * @param headHash - the current head of the chain (if null, do not update
     * `_headHeaderHash` and `_headBlockHash`)
     * @param ops - the `DatabaseOperation` list to add the delete operations to
     * @hidden
     */
    async _delChild(blockHash, blockNumber, headHash, ops) {
        var _a, _b;
        // delete header, body, hash to number mapping and td
        ops.push(helpers_1.DBOp.del(operation_1.DBTarget.Header, { blockHash, blockNumber }));
        ops.push(helpers_1.DBOp.del(operation_1.DBTarget.Body, { blockHash, blockNumber }));
        ops.push(helpers_1.DBOp.del(operation_1.DBTarget.HashToNumber, { blockHash }));
        ops.push(helpers_1.DBOp.del(operation_1.DBTarget.TotalDifficulty, { blockHash, blockNumber }));
        if (!headHash) {
            return;
        }
        if ((_a = this._headHeaderHash) === null || _a === void 0 ? void 0 : _a.equals(blockHash)) {
            this._headHeaderHash = headHash;
        }
        if ((_b = this._headBlockHash) === null || _b === void 0 ? void 0 : _b.equals(blockHash)) {
            this._headBlockHash = headHash;
        }
        try {
            const childHeader = await this._getCanonicalHeader(blockNumber.addn(1));
            await this._delChild(childHeader.hash(), childHeader.number, headHash, ops);
        }
        catch (error) {
            if (error.type !== 'NotFoundError') {
                throw error;
            }
        }
    }
    /**
     * Iterates through blocks starting at the specified iterator head and calls
     * the onBlock function on each block. The current location of an iterator
     * head can be retrieved using the `getHead()` method.
     *
     * @param name - Name of the state root head
     * @param onBlock - Function called on each block with params (block, reorg)
     */
    async iterator(name, onBlock) {
        return this._iterator(name, onBlock);
    }
    /**
     * @hidden
     */
    async _iterator(name, onBlock) {
        await this.initAndLock(async () => {
            const blockHash = this._heads[name] || this._genesis;
            let lastBlock;
            if (!blockHash) {
                return;
            }
            const number = await this.dbManager.hashToNumber(blockHash);
            const blockNumber = number.addn(1);
            // eslint-disable-next-line no-constant-condition
            while (true) {
                try {
                    const block = await this._getBlock(blockNumber);
                    if (!block) {
                        break;
                    }
                    this._heads[name] = block.hash();
                    const reorg = lastBlock ? lastBlock.hash().equals(block.header.parentHash) : false;
                    lastBlock = block;
                    await onBlock(block, reorg);
                    blockNumber.iaddn(1);
                }
                catch (error) {
                    if (error.type !== 'NotFoundError') {
                        throw error;
                    }
                    break;
                }
            }
            await this._saveHeads();
        });
    }
    /* Methods regarding re-org operations */
    /**
     * Pushes DB operations to delete canonical number assignments for specified
     * block number and above This only deletes `NumberToHash` references, and not
     * the blocks themselves. Note: this does not write to the DB but only pushes
     * to a DB operations list.
     * @param blockNumber - the block number from which we start deleting
     * canonical chain assignments (including this block)
     * @param headHash - the hash of the current canonical chain head. The _heads
     * reference matching any hash of any of the deleted blocks will be set to
     * this
     * @param ops - the DatabaseOperation list to write DatabaseOperations to
     * @hidden
     */
    async _deleteCanonicalChainReferences(blockNumber, headHash, ops) {
        var _a;
        blockNumber = blockNumber.clone();
        let hash;
        hash = await this.safeNumberToHash(blockNumber);
        while (hash) {
            ops.push(helpers_1.DBOp.del(operation_1.DBTarget.NumberToHash, { blockNumber }));
            // reset stale iterator heads to current canonical head this can, for
            // instance, make the VM run "older" (i.e. lower number blocks than last
            // executed block) blocks to verify the chain up to the current, actual,
            // head.
            Object.keys(this._heads).forEach((name) => {
                if (this._heads[name].equals(hash)) {
                    // explicitly cast as Buffer: it is not possible that `hash` is false
                    // here, but TypeScript does not understand this.
                    this._heads[name] = headHash;
                }
            });
            // reset stale headBlock to current canonical
            if ((_a = this._headBlockHash) === null || _a === void 0 ? void 0 : _a.equals(hash)) {
                this._headBlockHash = headHash;
            }
            blockNumber.iaddn(1);
            hash = await this.safeNumberToHash(blockNumber);
        }
    }
    /**
     * Given a `header`, put all operations to change the canonical chain directly
     * into `ops`. This walks the supplied `header` backwards. It is thus assumed
     * that this header should be canonical header. For each header the
     * corresponding hash corresponding to the current canonical chain in the DB
     * is checked If the number => hash reference does not correspond to the
     * reference in the DB, we overwrite this reference with the implied number =>
     * hash reference Also, each `_heads` member is checked; if these point to a
     * stale hash, then the hash which we terminate the loop (i.e. the first hash
     * which matches the number => hash of the implied chain) is put as this stale
     * head hash The same happens to _headBlockHash
     * @param header - The canonical header.
     * @param ops - The database operations list.
     * @hidden
     */
    async _rebuildCanonical(header, ops) {
        var _a;
        const currentNumber = header.number.clone(); // we change this during this method with `isubn`
        let currentCanonicalHash = header.hash();
        // track the staleHash: this is the hash currently in the DB which matches
        // the block number of the provided header.
        let staleHash = false;
        let staleHeads = [];
        let staleHeadBlock = false;
        const loopCondition = async () => {
            staleHash = await this.safeNumberToHash(currentNumber);
            currentCanonicalHash = header.hash();
            return !staleHash || !currentCanonicalHash.equals(staleHash);
        };
        while (await loopCondition()) {
            // handle genesis block
            const blockHash = header.hash();
            const blockNumber = header.number;
            helpers_1.DBSaveLookups(blockHash, blockNumber).map((op) => {
                ops.push(op);
            });
            if (blockNumber.isZero()) {
                break;
            }
            // mark each key `_heads` which is currently set to the hash in the DB as
            // stale to overwrite this later.
            Object.keys(this._heads).forEach((name) => {
                if (staleHash && this._heads[name].equals(staleHash)) {
                    staleHeads.push(name);
                }
            });
            // flag stale headBlock for reset
            if (staleHash && ((_a = this._headBlockHash) === null || _a === void 0 ? void 0 : _a.equals(staleHash))) {
                staleHeadBlock = true;
            }
            currentNumber.isubn(1);
            try {
                header = await this._getHeader(header.parentHash, currentNumber);
            }
            catch (error) {
                staleHeads = [];
                if (error.type !== 'NotFoundError') {
                    throw error;
                }
                break;
            }
        }
        // the stale hash is equal to the blockHash set stale heads to last
        // previously valid canonical block
        staleHeads.forEach((name) => {
            this._heads[name] = currentCanonicalHash;
        });
        // set stale headBlock to last previously valid canonical block
        if (staleHeadBlock) {
            this._headBlockHash = currentCanonicalHash;
        }
    }
    /* Helper functions */
    /**
     * Builds the `DatabaseOperation[]` list which describes the DB operations to
     * write the heads, head header hash and the head header block to the DB
     * @hidden
     */
    _saveHeadOps() {
        return [
            helpers_1.DBOp.set(operation_1.DBTarget.Heads, this._heads),
            helpers_1.DBOp.set(operation_1.DBTarget.HeadHeader, this._headHeaderHash),
            helpers_1.DBOp.set(operation_1.DBTarget.HeadBlock, this._headBlockHash),
        ];
    }
    /**
     * Gets the `DatabaseOperation[]` list to save `_heads`, `_headHeaderHash` and
     * `_headBlockHash` and writes these to the DB
     * @hidden
     */
    async _saveHeads() {
        return this.dbManager.batch(this._saveHeadOps());
    }
    /**
     * Gets a header by hash and number. Header can exist outside the canonical
     * chain
     *
     * @hidden
     */
    async _getHeader(hash, number) {
        if (!number) {
            number = await this.dbManager.hashToNumber(hash);
        }
        return this.dbManager.getHeader(hash, number);
    }
    /**
     * Gets a header by number. Header must be in the canonical chain
     *
     * @hidden
     */
    async _getCanonicalHeader(number) {
        const hash = await this.dbManager.numberToHash(number);
        return this._getHeader(hash, number);
    }
    /**
     * This method either returns a Buffer if there exists one in the DB or if it
     * does not exist (DB throws a `NotFoundError`) then return false If DB throws
     * any other error, this function throws.
     * @param number
     */
    async safeNumberToHash(number) {
        try {
            const hash = await this.dbManager.numberToHash(number);
            return hash;
        }
        catch (error) {
            if (error.type !== 'NotFoundError') {
                throw error;
            }
            return false;
        }
    }
}
exports.default = Blockchain;
//# sourceMappingURL=index.js.map